static	char	const	rcsid[] = "$Id: clamav-milter.c,v 1.18 2003/11/11 08:19:20 nigelhorne Exp $";

#define	CM_VERSION	"w1"

#include "defaults.h"
#include "cfgfile.h"
#include "../target.h"

#include <stdio.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdarg.h>
#include <errno.h>
#include <libmilter/mfapi.h>
#include <pthread.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <signal.h>
#include <regex.h>
#include <fcntl.h>

#define _GNU_SOURCE
#include "getopt.h"

struct	privdata {
	char	*from;
	char	**to;
	int	numTo;
	int	cmdSocket;
	int	dataSocket;
	char	**buffer;
	int	size;
	int	numBuf;
	int	fadd;
};

static	int	pingServer(void);
static	sfsistat	clamfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr);
static	sfsistat	clamfi_envfrom(SMFICTX *ctx, char **argv);
static	sfsistat	clamfi_envrcpt(SMFICTX *ctx, char **argv);
static	sfsistat	clamfi_header(SMFICTX *ctx, char *headerf, char *headerv);
static	sfsistat	clamfi_eoh(SMFICTX *ctx);
static	sfsistat	clamfi_body(SMFICTX *ctx, u_char *bodyp, size_t len);
static	sfsistat	clamfi_eom(SMFICTX *ctx);
static	sfsistat	clamfi_abort(SMFICTX *ctx);
static	sfsistat	clamfi_close(SMFICTX *ctx);
static	void		clamfi_cleanup(SMFICTX *ctx);
static	int		clamfi_send(struct privdata *privdata, size_t len, char *format, ...);
static	int		sendbuf(struct privdata *privdata, size_t len, char *ptr);
static	char		*strrcpy(char *dest, const char *source);
static int		conn(struct privdata *privdata);

static	char	clamav_version[128];
static	int	cl_error = SMFIS_TEMPFAIL;

static	int	use_syslog = 0;
static	int	logVerbose = 0;
static	struct	cfgstruct	*copt;
static	const	char	*localSocket;
static	in_port_t	tcpSocket;
static	const	char	*serverIP = "127.0.0.1";

static void
help(void)
{
	printf("\n\tclamav-milter version %s\n", CM_VERSION);
}

int
main(int argc, char **argv)
{
	extern char *optarg;
	char *port = NULL;
	const char *cfgfile = CL_DEFAULT_CFG;
	struct cfgstruct *cpt;
	struct smfiDesc smfilter = {
		"ClamAv",
		SMFI_VERSION,
		SMFIF_ADDHDRS,
		clamfi_connect,
		NULL,
		clamfi_envfrom,
		clamfi_envrcpt,
		clamfi_header,
		clamfi_eoh,
		clamfi_body,
		clamfi_eom,
		clamfi_abort,
		clamfi_close,
	};

	snprintf(clamav_version, sizeof(clamav_version),
		"ClamAV version %s, clamav-milter version %s",
		VERSION, CM_VERSION);

	for(;;) {
		int opt_index = 0;
		const char *args = "bc:flnopPqdhs:V";
		static struct option long_options[] = {
			{
				"config-file", 1, NULL, 'c'
			},
			{
				"help", 0, NULL, 'h'
			},
			{
				"local", 0, NULL, 'l'
			},
			{
				"server", 1, NULL, 's'
			},
			{
				"version", 0, NULL, 'V'
			},
			{
				NULL, 0, NULL, '\0'
			}
		};

		int ret = getopt_long(argc, argv, args, long_options, &opt_index);

		if(ret == -1)
			break;
		else if(ret == 0)
			ret = long_options[opt_index].val;

		switch(ret) {
			case 'c':
				cfgfile = optarg;
				break;
			case 'h':
				help();
				return EX_OK;
			case 's':
				serverIP = optarg;
				break;
			case 'v':
				puts(clamav_version);
				return EX_OK;
			default:
				fprintf(stderr, "Usage: %s [-c=FILE] socket-addr\n", argv[0]);
				return EX_USAGE;
		}
	}

	if (optind == argc) {
		fprintf(stderr, "%s: No socket-addr given\n", argv[0]);
		return EX_USAGE;
	}
	port = argv[optind];

	if((copt = parsecfg(cfgfile)) == NULL) {
		fprintf(stderr, "%s: Can't parse the config file %s\n",
			argv[0], cfgfile);
		return EX_CONFIG;
	}

	if(!cfgopt(copt, "StreamSaveToDisk")) {
		fprintf(stderr, "%s: StreamSavetoDisk not enabled in %s\n",
			argv[0], cfgfile);
		return EX_CONFIG;
	}

	if(!cfgopt(copt, "ScanMail")) {
		fprintf(stderr, "%s: ScanMail not enabled in %s\n",
			argv[0], cfgfile);
		return EX_CONFIG;
	}

	if((cpt = cfgopt(copt, "LocalSocket")) != NULL) {
		if(cfgopt(copt, "TCPSocket") != NULL) {
			fprintf(stderr, "%s: You can select one server type only (local/TCP) in %s\n",
				argv[0], cfgfile);
			return EX_CONFIG;
		}
		localSocket = cpt->strarg;
		if(!pingServer()) {
			fprintf(stderr, "Can't talk to clamd server via %s\n",
				localSocket);
			fprintf(stderr, "Check your entry for LocalSocket in %s\n",
				cfgfile);
			return EX_CONFIG;
		}
	} else if((cpt = cfgopt(copt, "TCPSocket")) != NULL) {
		tcpSocket = (in_port_t)cpt->numarg;
		if(!pingServer()) {
			fprintf(stderr, "Can't talk to clamd server at %s on port %d\n",
				serverIP, tcpSocket);
			fprintf(stderr, "Check your entry for TCPSocket in %s\n",
				cfgfile);
			return EX_CONFIG;
		}
	} else {
		fprintf(stderr, "%s: You must select server type (local/TCP) in %s\n",
			argv[0], cfgfile);
		return EX_CONFIG;
	}

	switch(fork()) {
		case -1:
			perror("fork");
			return EX_TEMPFAIL;
		case 0:
			break;
		default:
			return EX_OK;
	}
	close(0);
	close(1);
	close(2);
	open("/dev/null", O_RDONLY);
	if(open("/dev/console", O_WRONLY) == 1)
		dup(1);

	if(smfi_setconn(port) == MI_FAILURE) {
		fprintf(stderr, "%s: smfi_setconn failed\n",
			argv[0]);
		return EX_SOFTWARE;
	}

	if(cfgopt(copt, "LogSyslog")) {
		openlog("clamav-milter", LOG_CONS|LOG_PID, LOG_MAIL);
		syslog(LOG_INFO, clamav_version);
		use_syslog = 1;

		if(cfgopt(copt, "LogVerbose"))
			logVerbose = 1;
	} else {
		use_syslog = 0;
	}

	if(cfgopt(copt, "FixStaleSocket")) {
		if(strncasecmp(port, "unix:", 5) == 0) {
			if(unlink(&port[5]) < 0)
				perror(&port[5]);
		} else if(strncasecmp(port, "local:", 6) == 0) {
			if(unlink(&port[6]) < 0)
				perror(&port[6]);
		}
	}

	if(smfi_register(smfilter) == MI_FAILURE) {
		fputs("smfi_register failure\n", stderr);
		return EX_UNAVAILABLE;
	}

	signal(SIGPIPE, SIG_IGN);

	return smfi_main();
}

static int
pingServer(void)
{
	char *ptr;
	int sock, nbytes;
	char buf[128];

	if(localSocket) {
		struct sockaddr_un server;

		memset((char *)&server, 0, sizeof(struct sockaddr_un));
		server.sun_family = AF_UNIX;
		strncpy(server.sun_path, localSocket, sizeof(server.sun_path));

		if((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
			perror("socket");
			return 0;
		}
		if(connect(sock, (struct sockaddr *)&server, sizeof(struct sockaddr_un)) < 0) {
			perror(localSocket);
			return 0;
		}
	} else {
		struct sockaddr_in server;

		memset((char *)&server, 0, sizeof(struct sockaddr_in));
		server.sin_family = AF_INET;
		server.sin_port = htons(tcpSocket);
		server.sin_addr.s_addr = inet_addr(serverIP);

		if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			perror("socket");
			return 0;
		}
		if(connect(sock, (struct sockaddr *)&server, sizeof(struct sockaddr_in)) < 0) {
			perror("connect");
			return 0;
		}
	}

	if(send(sock, "VERSION\n", 8, 0) < 8) {
		perror("send");
		close(sock);
		return 0;
	}

	shutdown(sock, SHUT_WR);

	nbytes = recv(sock, buf, sizeof(buf), 0);

	close(sock);

	if(nbytes < 0) {
		perror("recv");
		return 0;
	}
	buf[nbytes] = '\0';

	if((ptr = strchr(buf, '\n')) != NULL)
		*ptr = '\0';

	snprintf(clamav_version, sizeof(clamav_version),
		"ClamAV version '%s', clamav-milter version '%s'",
		buf, CM_VERSION);
		
	return 1;
}

static sfsistat
clamfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	char buf[INET_ADDRSTRLEN];

	if(use_syslog)
	{
		const char *remoteIP = inet_ntop(AF_INET, &((struct sockaddr_in *)(hostaddr))->sin_addr, buf, sizeof(buf));
		syslog(LOG_NOTICE, "clamfi_connect: connection from %s [%s]", hostname, remoteIP);
	}

	return SMFIS_CONTINUE;
}

static sfsistat
clamfi_envfrom(SMFICTX *ctx, char **argv)
{
	struct privdata *privdata;

	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_envfrom: %s", argv[0]);

	privdata = (struct privdata *)malloc(sizeof(struct privdata));
	privdata->dataSocket = -1;
	privdata->cmdSocket = -1;
	privdata->numTo = 0;
	privdata->buffer = (char**)calloc(sizeof(char*),100);
	privdata->numBuf = 0;
	privdata->size = 0;
	privdata->fadd = 1;

	privdata->buffer[0] = (char*)malloc(10000);

	clamfi_send(privdata, 0, "From %s\n", argv[0]);
	clamfi_send(privdata, 0, "From: %s\n", argv[0]);

	privdata->from = strdup(argv[0]);
	privdata->to = NULL;

	return (smfi_setpriv(ctx, privdata) == MI_SUCCESS) ? SMFIS_CONTINUE : cl_error;
}

static sfsistat
clamfi_envrcpt(SMFICTX *ctx, char **argv)
{
	struct privdata *privdata = (struct privdata *)smfi_getpriv(ctx);

	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_envrcpt: %s", argv[0]);

	clamfi_send(privdata, 0, "To: %s\n", argv[0]);

	if(privdata->to == NULL) {
		privdata->to = malloc(sizeof(char *) * 2);
	} else
		privdata->to = realloc(privdata->to, sizeof(char *) * (privdata->numTo + 2));

	privdata->to[privdata->numTo] = strdup(argv[0]);
	privdata->to[++privdata->numTo] = NULL;

	return SMFIS_CONTINUE;
}

static sfsistat
clamfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	struct privdata *privdata = (struct privdata *)smfi_getpriv(ctx);

	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_header: %s: %s", headerf, headerv);

	if(clamfi_send(privdata, 0, "%s: %s\n", headerf, headerv) < 0) {
		clamfi_cleanup(ctx);
		return cl_error;
	}
	return SMFIS_CONTINUE;
}

static sfsistat
clamfi_eoh(SMFICTX *ctx)
{
	struct privdata *privdata = (struct privdata *)smfi_getpriv(ctx);
	char **to;

	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_eoh");

	if(clamfi_send(privdata, 1, "\n") < 0) {
		clamfi_cleanup(ctx);
		return cl_error;
	}

	return SMFIS_CONTINUE;
}

static sfsistat
clamfi_body(SMFICTX *ctx, u_char *bodyp, size_t len)
{
	struct privdata *privdata = (struct privdata *)smfi_getpriv(ctx);

	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_envbody: %u bytes", len);

	if(clamfi_send(privdata, len, (char *)bodyp) < 0) {
		clamfi_cleanup(ctx);
		return cl_error;
	}
	return SMFIS_CONTINUE;
}

static sfsistat
clamfi_eom(SMFICTX *ctx)
{
	int rc = SMFIS_CONTINUE;
	char *ptr;
	struct privdata *privdata = (struct privdata *)smfi_getpriv(ctx);
	char mess[128];

	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_eom");

	if(privdata->fadd) {
		int i;
		privdata->fadd = 0;
		if(conn(privdata)) {
		    clamfi_cleanup(ctx);
		    return cl_error;
		}
		for(i=0;i<100;i++)
		    if(privdata->buffer[i])
			if(sendbuf(privdata, 10000, privdata->buffer[i])) {
			    clamfi_cleanup(ctx);
			    return cl_error;
			}
	}

	close(privdata->dataSocket);
	privdata->dataSocket = -1;

	if(recv(privdata->cmdSocket, mess, sizeof(mess), 0) > 0) {
		if((ptr = strchr(mess, '\n')) != NULL)
			*ptr = '\0';

		if(logVerbose)
			syslog(LOG_DEBUG, "clamfi_eom: read %s", mess);
	} else {
		syslog(LOG_NOTICE, "clamfi_eom: read nothing from clamd");
		mess[0] = '\0';
	}

	if(strstr(mess, "FOUND") == NULL) {
		smfi_addheader(ctx, "X-Virus-Scanned", clamav_version);

		if(use_syslog)
			syslog(LOG_NOTICE, "clean message from %s",
				(privdata->from) ? privdata->from : "an unknown sender");
	} else {
		int i;
		char **to, *err;
		FILE *sendmail;

		if(use_syslog)
			syslog(LOG_NOTICE, mess);

		err = (char *)malloc(1024);

		snprintf(err, 1024, "Intercepted virus from %s to", privdata->from);

		ptr = strchr(err, '\0');

		i = 1024;

		for(to = privdata->to; *to; to++) {
			if(&ptr[strlen(*to) + 2] >= &err[i]) {
				i += 1024;
				err = realloc(err, i);
			}
			ptr = strrcpy(ptr, " ");
			ptr = strrcpy(ptr, *to);
		}
		(void)strcpy(ptr, "\n");

		if(use_syslog)
			syslog(LOG_NOTICE, "%s", err);

		smfi_setreply(ctx, "550", "5.7.1", "Virus detected by ClamAV - http://clamav.elektrapro.com");
		rc = SMFIS_REJECT;
		free(err);
	}
	clamfi_cleanup(ctx);

	return rc;
}

static sfsistat
clamfi_abort(SMFICTX *ctx)
{
	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_abort");

	clamfi_cleanup(ctx);

	return cl_error;
}

static sfsistat
clamfi_close(SMFICTX *ctx)
{
	if(logVerbose)
		syslog(LOG_DEBUG, "clamfi_close");

	clamfi_cleanup(ctx);

	return SMFIS_CONTINUE;
}

static void
clamfi_cleanup(SMFICTX *ctx)
{
	struct privdata *privdata = (struct privdata *)smfi_getpriv(ctx);

	if(privdata) {
		if(privdata->dataSocket >= 0) {
			close(privdata->dataSocket);
			privdata->dataSocket = -1;
		}

		if(privdata->from) {
			free(privdata->from);
			privdata->from = NULL;
		}

		if(privdata->to) {
			char **to;

			for(to = privdata->to; *to; to++) {
				free(*to);
			}
			free(privdata->to);
			privdata->to = NULL;
		}

		if(privdata->cmdSocket >= 0) {
			char buf[64];

			while(recv(privdata->cmdSocket, buf, sizeof(buf), 0) > 0)
				;
			close(privdata->cmdSocket);
			privdata->cmdSocket = -1;
		}

		if(privdata->buffer) {
			int i;
			for(i=0;i<100;i++)
				if(privdata->buffer[i])
					free(privdata->buffer[i]);
			free(privdata->buffer);
		}

		free(privdata);
		smfi_setpriv(ctx, NULL);
	}
}

static int
clamfi_send(struct privdata *privdata, size_t len, char *format, ...)
{
	char output[BUFSIZ];
	char *ptr;
	int rc = 0;

	if(len > 0)
		ptr = format;
	else {
		va_list argp;

		va_start(argp, format);
		vsnprintf(output, sizeof(output), format, argp);
		va_end(argp);

		len = strlen(output);
		ptr = output;
	}
	
	if(!privdata->fadd)
		return sendbuf(privdata, len, ptr);
	else {
		int n = 10000-privdata->size;
		if(len<n) {
			memcpy(&privdata->buffer[privdata->numBuf][privdata->size],ptr,len);
			privdata->size += len;
			return 0;
		} else {
			memcpy(&privdata->buffer[privdata->numBuf++][privdata->size],ptr,n);
			len -= n;
			ptr = &ptr[n];
			while((len>=10000)&&(privdata->numBuf<100)) {
				privdata->buffer[privdata->numBuf] = (char *)malloc(10000);
				memcpy(privdata->buffer[privdata->numBuf++],ptr,n);
				len -= 10000;
				ptr = &ptr[10000];
			}
			if(privdata->numBuf==100) {
				privdata->fadd = 0;
				if(conn(privdata))
				    return -1;
				for(n=0;n<100;n++)
					rc |= sendbuf(privdata, 10000, privdata->buffer[n]);
				if(len)
					rc |= sendbuf(privdata, len, ptr);
			} else {
				privdata->buffer[privdata->numBuf] = (char *)malloc(10000);
				memcpy(privdata->buffer[privdata->numBuf],ptr,len);
				privdata->size = len;
			}
		}
	}

	return rc;
}

static char *
strrcpy(char *dest, const char *source)
{
	while((*dest++ = *source++) != '\0')
		;
	return(--dest);
}

static int
sendbuf(struct privdata *privdata, size_t len, char *ptr)
{
	while(len > 0) {
		int nbytes = send(privdata->dataSocket, ptr, len, 0);

		if(nbytes == -1) {
			if(errno == EINTR)
				continue;
			perror("send");
			if(use_syslog)
				syslog(LOG_ERR, "write failure to clamd");

			return -1;
		}
		len -= nbytes;
		ptr = &ptr[nbytes];
	}
	return 0;
}

static int
conn(struct privdata *privdata)
{
	struct sockaddr_in reply;
	unsigned short port;
	int nbytes, rc;
	char buf[64];

	if(localSocket) {
		struct sockaddr_un server;

		memset((char *)&server, 0, sizeof(struct sockaddr_un));
		server.sun_family = AF_UNIX;
		strncpy(server.sun_path, localSocket, sizeof(server.sun_path));

		if((privdata->cmdSocket = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
			perror("socket");
			return -1;
		}
		if(connect(privdata->cmdSocket, (struct sockaddr *)&server, sizeof(struct sockaddr_un)) < 0) {
			perror(localSocket);
			return -1;
		}
	} else {
		struct sockaddr_in server;

		memset((char *)&server, 0, sizeof(struct sockaddr_in));
		server.sin_family = AF_INET;
		server.sin_port = htons(tcpSocket);
		server.sin_addr.s_addr = inet_addr(serverIP);

		if((privdata->cmdSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			perror("socket");
			return -1;
		}
		if(connect(privdata->cmdSocket, (struct sockaddr *)&server, sizeof(struct sockaddr_in)) < 0) {
			perror("connect");
			return -1;
		}
	}

	if((privdata->dataSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		close(privdata->cmdSocket);
		free(privdata);
		if(use_syslog)
			syslog(LOG_ERR, "failed to create socket");
		return -1;
	}

	shutdown(privdata->dataSocket, SHUT_RD);

	if(send(privdata->cmdSocket, "STREAM\n", 7, 0) < 7) {
		perror("send");
		close(privdata->dataSocket);
		close(privdata->cmdSocket);
		free(privdata);
		if(use_syslog)
			syslog(LOG_ERR, "send failed to clamd");
		return -1;
	}

	shutdown(privdata->cmdSocket, SHUT_WR);

	nbytes = recv(privdata->cmdSocket, buf, sizeof(buf), 0);
	if(nbytes < 0) {
		perror("recv");
		close(privdata->dataSocket);
		close(privdata->cmdSocket);
		free(privdata);
		if(use_syslog)
			syslog(LOG_ERR, "recv failed from clamd getting PORT");
		return -1;
	}
	buf[nbytes] = '\0';
	if(sscanf(buf, "PORT %hu\n", &port) != 1) {
		close(privdata->dataSocket);
		close(privdata->cmdSocket);
		free(privdata);
		if(use_syslog)
			syslog(LOG_ERR, "Expected port information from clamd, got '%s'",
				buf);
		else
			fprintf(stderr, "Expected port information from clamd, got '%s'\n",
				buf);
		return -1;
	}

	memset((char *)&reply, 0, sizeof(struct sockaddr_in));
	reply.sin_family = AF_INET;
	reply.sin_port = ntohs(port);

	reply.sin_addr.s_addr = inet_addr(serverIP);

	rc = connect(privdata->dataSocket, (struct sockaddr *)&reply, sizeof(struct sockaddr_in));

	if(rc < 0) {
		perror("connect");

		close(privdata->dataSocket);
		close(privdata->cmdSocket);
		free(privdata);

		if(use_syslog) {
			strerror_r(rc, buf, sizeof(buf));
			syslog(LOG_ERR, "Failed to connect to port %d given by clamd: %s", port, buf);
		}

		return -1;
	}
	return 0;
}
